#!/usr/bin/perl -w
#
# Parse out Netscape/Mozilla or Opera bookmarks and build a XML output
# file, suitable for importing into HTBM
#
# Usage:
# ./parse_bookmarks.pl [-m] [-o] [input-file] [output-file]
#    -m     Parse as Netscape/Mozilla
#    -o     Parse as Opera
# (one && only one required)

use strict;
use Fcntl ":mode";
use HTML::Entities;

# Parse command-line args
if(@ARGV != 3)
{
	print(STDERR "Error: Need 3 arguments.\n");
	usage();
}

# What kinda bookmarks are we dealing with?
my $mode = $ARGV[0];
if( !defined($mode) || (($mode ne "-m") && ($mode ne "-o")) )
{
	print(STDERR "Error: Unknown mode '$mode'.\n");
	usage();
}
if($mode eq "-m")
{ $mode = "mozilla"; }
elsif($mode eq "-o")
{ $mode = "opera"; }

# Get our input file
my $input = $ARGV[1];
if(!defined($input))
{
	# We shouldn't be able to get here ever
	print(STDERR "Error: No input file specified.\n");
	usage();
}
# Allow standard alias '-' for stdin
if($input ne "-")
{
	my @stret = stat($input);
	if(!@stret)
	{
		print(STDERR "Error: Input file '$input' doesn't seem to exist.\n");
		usage();
	}
	if(!(-r _))
	{
		print(STDERR "Error: I can't read input file '$input'.\n");
		usage();
	}
	my $fmode = $stret[2];
	if( !($fmode & S_IFREG) )
	{
		print(STDERR "Error: Input file '$input' doesn't seem like a regular file.\n");
		usage();
	}
}

# And our output file
my $output = $ARGV[2];
if(!defined($output))
{
	# We shouldn't be able to get here ever
	print(STDERR "Error: No output file specified.\n");
	usage();
}
# Allow standard alias '-' for stdout
if($output ne "-")
{
	my @stret = stat($output);
	if(@stret)
	{
		print(STDERR "Error: Output file '$output' already exists!\n");
		usage();
	}
}


#
# Now we can get into the meat.
#
my $xml_categories = "";
my $xml_bookmarks = "";

# Open our input file
if($input eq "-")
{ open(INFILE, "<&STDIN") or die "Can't open input file: $!"; }
else
{ open(INFILE, "<$input") or die "Can't open input file: $!"; }
my @inp = <INFILE>;
close(INFILE);

# Start processing
my @catpath = (0);
my $curcat = 0;
my $nextcat = 1;
my $parentcat = 0;
my $curbm;
my $nextbm = 1;
my $i;
if($mode eq "mozilla")
{
	# Mozilla format
	#
	# Mozilla (and pre-Mozilla Netscape too) uses an
	# almost-pseudo-HTML-ish format for its bookmark files, based in a
	# <dl>, so we have to untangle it.  Also, look out for characters
	# that were made HTML entities; we have to decode them, THEN add back
	# in the entities we need for XML.
	for($i=0 ; $i < @inp ; $i++) {
		# Get&trim
		$_ = $inp[$i]; chomp; s/\s+$//; s/^\s+//;

		# Now, what's on this line?
		if(/^<DT><H3/i)
		{
			# We're starting a new category
			$curcat = $nextcat;
			$nextcat++;

			push(@catpath, $parentcat);

			# Find the name
			my ($cname) = />([^<]+)<\/H3>$/i;
			# De-HTML-entitize and XML-entitize
			$cname = decode_entities($cname);
			$cname =~ s/&/&amp;/g;
			$cname =~ s/</&lt;/g;
			# It's in an attribute
			$cname =~ s/"/&quot;/g;

			# Debug
			#print(STDERR "Got cname '$cname'\n");


			# Append into our list
			$xml_categories .= "  <category id=\"$curcat\" "
					. "name=\"$cname\">\n"
					. "   <parent id=\"$parentcat\" />\n"
					. "  </category>\n";

			# Now update the parent category, since we're it from now on
			$parentcat = $curcat;

			# We aren't using longer descriptions of categories, so skip
			# one if it's there
			if($inp[$i+1] =~ /^\s*<DD>/i)
			{ $i++; }
		}
		elsif(/^<\/DL/i)
		{
			# End of category
			$parentcat = pop(@catpath);
		}
		elsif(/^<DT><A HREF/i)
		{
			# A bookmark!  Joy!
			$curbm = $nextbm;
			$nextbm++;

			# Get the URL
			my ($burl) = /<A HREF="([^"]+)"/i;
			# Fixup for XML entities
			$burl =~ s/&/&amp;/g;

			# Get the name
			my ($bname) = />([^<]+)<\/A>$/i;
			# De-HTML-entitize and XML-entitize
			$bname = decode_entities($bname);
			$bname =~ s/&/&amp;/g;
			$bname =~ s/</&lt;/g;

			# Get the comment, if any
			my $comment = "";
			if($inp[$i+1] =~ /^\s*<DD/)
			{
				$i++;
				$_ = $inp[$i]; chomp; s/\s+$//; s/^\s+//;
				($comment) = /<DD>(.+)/i;

				# This can go on for multiple lines
				while($inp[$i+1] !~ /^\s*</)
				{
					$i++;
					$_ = $inp[$i]; chomp; s/\s+$//; s/^\s+//;

					$comment .= "\n$_";
				}
			}
			# De-HTML-entitize and XML-entitize
			$comment = decode_entities($comment);
			$comment =~ s/&/&amp;/g;
			$comment =~ s/</&lt;/g;

			# Debug
			#print(STDERR "Got bookmark '$bname' ($burl)'\n");


			# Append into our list
			$xml_bookmarks .= "  <bookmark id=\"$curbm\">\n"
					. "   <url>$burl</url>\n"
					. "   <title>$bname</title>\n"
					. "   <comment>$comment</comment>\n"
					. "   <created></created>\n"
					. "   <parent id=\"$parentcat\" />\n"
					. "  </bookmark>\n";

		}
	} # Loop over lines
}
elsif($mode eq "opera")
{
	# Opera format
	#
	# In many ways, this is easier than Mozilla's format, since there's
	# always a blank line separating anything.  Makes it much easier,
	# with less peaking and state-tracking.  It's also not pretending to
	# be HTML, so we don't have to worry about weird formatting or
	# entities.
	for($i=0 ; $i < @inp ; $i++)
	{
		$_ = $inp[$i]; chomp; s/\s+$//; s/^\s+//;

		# Any lines we care about start with '#' or '-'
		if(!(/^[#\-]/))
		{ next; }

		# Have we a category?
		if(/^#FOLDER/)
		{
			$curcat = $nextcat;
			$nextcat++;

			push(@catpath, $parentcat);

			# Name is the only thing we care about in categories
			my $name;
			while(($_ = $inp[++$i]) !~ /^\s*$/)
			{
				chomp; s/\s+$//; s/^\s+//;
				if(/^NAME/)
				{
					($name) = /^NAME=(.*)/;
					# Entitize
					$name =~ s/&/&amp;/g;
					$name =~ s/</&lt;/g;
					$name =~ s/"/&quot;/g;
				}
				# We don't care about the other fields
			}

			# Create the entry
			# Append into our list
			$xml_categories .= "  <category id=\"$curcat\" "
					. "name=\"$name\">\n"
					. "   <parent id=\"$parentcat\" />\n"
					. "  </category>\n";

			# Now update the parent category, since we're it from now on
			$parentcat = $curcat;
		}
		elsif(/^-$/)
		{
			# Back up a category
			$parentcat = pop(@catpath);
		}
		elsif(/^#URL/)
		{
			# A bookmark
			$curbm = $nextbm;
			$nextbm++;
			my $url;
			my $title;
			# Is there a comment field?  I'm not sure...
			while(($_ = $inp[++$i]) !~ /^\s*$/)
			{
				chomp; s/\s+$//; s/^\s+//;
				if(/^NAME/)
				{
					($title) = /^NAME=(.*)/;
					# Entitize
					$title =~ s/&/&amp;/g;
					$title =~ s/</&lt;/g;
				}
				elsif(/^URL/)
				{
					($url) = /^URL=(.*)/;
					# Entitize
					# Opera does NOT pre-entitize anything
					$url =~ s/&/&amp;/g;
					$url =~ s/</&lt;/g;
				}
				# We don't care about other fields
			}

			# Add it on
			$xml_bookmarks .= "  <bookmark id=\"$curbm\">\n"
					. "   <url>$url</url>\n"
					. "   <title>$title</title>\n"
					. "   <comment></comment>\n"
					. "   <created></created>\n"
					. "   <parent id=\"$parentcat\" />\n"
					. "  </bookmark>\n";
		} # Check lines
	} # Loop over lines
} # Opera mode


# Now do the output
if($output eq "-")
{ open(OUTFILE, ">&STDOUT") or die "Can't open output file: $!"; }
else
{ open(OUTFILE, ">$output") or die "Can't open output file: $!"; }

print OUTFILE << "_EOF_";
<?xml version="1.0" encoding="iso-8859-1"?>
<!DOCTYPE htbmbak SYSTEM "htbmbak.dtd">
<htbmbak version="1.0">
 <categories>
  <category id="0" name="Bookmarks">
  </category>
$xml_categories </categories>
 <bookmarks>
$xml_bookmarks </bookmarks>
</htbmbak>
_EOF_

close(OUTFILE);

# Fin
exit(0);

sub usage
{
	print(STDERR "Usage: $0 [mode] [input-file] [output-file]\n"
			. "\t [mode] = -m (Mozilla) or -o (Opera)\n"
			. "\t [input-file] is the browser's bookmark file\n"
			. "\t [output-file] will be where the XML output will be "
			. "written (must not already exist).\n");
	exit(1);
}
